import os

from click import Command
from flask import Blueprint
from flask_sqlalchemy import DefaultMeta

env = os.environ

DBModel = DefaultMeta
Router = Blueprint
AppCommand = Command

# Specify name of the module the application should look for.
# String and array of strings are accepted.
# You can use a module path, relative to root of the sub-app.
# Example: "models", "sub_app.models" or ["sub_app.sqla_models", "sub_app.models.sqla_models"]
MODELS_FILE = env.get("FLASK_SUB_APPS_MODELS_FILE", ["models", "sqla_models"])
COMMANDS_FILE = env.get("FLASK_SUB_APPS_COMMANDS_FILE", ["commands", "flask_commands"])
ROUTERS_FILE = env.get(
    "FLASK_SUB_APPS_ROUTERS_FILE",
    ["routers", "views", "api_views", "urls", "blueprints"],
)



from typing import Optional, Type, Dict, List

from flask import Flask

from flask_sub_apps import helpers, config
from flask_sub_apps.loader import ModuleLoader, ValidateFunction
from flask_sub_apps.models import SubApp

logger = helpers.get_logger()


class AppRegistry:
    """Registry of Flask sub-apps."""

    sub_apps: Dict[str, SubApp] = dict()

    def __init__(
        self,
        is_valid_model: ValidateFunction = None,
        is_valid_command: ValidateFunction = None,
        is_valid_router: ValidateFunction = None,
    ):
        self._loader = ModuleLoader(is_valid_model, is_valid_command, is_valid_router)

    def get_sub_apps(self) -> List[SubApp]:
        """
        Get all registered sub-apps.
        :return: List of registered sub-apps.
        """
        return list(self.sub_apps.values())

    def register(self, module_name: str, app: Flask) -> Optional[SubApp]:
        """
        Register module as a Flask sub-app.

        :param module_name: module name.
        :param app: Flask app.
        :return: Sub-app with modules or None if failed to register.
        """
        try:
            sub_app = self._loader.load_sub_app(module_name)

            self.sub_apps[sub_app.module_name] = sub_app
            self.register_routers(app)
            self.register_commands(app)
            return sub_app
        except ModuleNotFoundError as e:
            logger.error(f"Failed to register module {module_name!r}", e)

    def register_routers(self, app: Flask):
        """Register Flask routers of all registered modules."""
        for router in self.get_routers():
            if app.blueprints.get(router.name):
                logger.debug(f"Router {router.name} already registered.")
                continue

            app.register_blueprint(router)
            logger.debug(f"Registered {router.name!r} blueprint")

    def register_commands(self, app: Flask):
        """Register Flask commands of all registered modules."""
        for command in self.get_commands():
            if app.cli.commands.get(command.name):
                logger.debug(f"Command {command.name} already registered.")
                continue

            app.cli.add_command(command, name=command.name)
            logger.debug(f"Registered {command.name!r} command")

    def get_routers(self) -> List[config.Router]:
        """Get all registered routers."""
        return [view for sub_app in self.get_sub_apps() for view in sub_app.routers]

    def get_commands(self) -> List[config.Command]:
        """Get all registered commands."""
        return [
            command for sub_app in self.get_sub_apps() for command in sub_app.commands
        ]

    def get_models(self) -> List[Type[config.DBModel]]:
        """Get all registered models."""
        return [model for sub_app in self.get_sub_apps() for model in sub_app.models]

    def get_model(self, model_name: str) -> Optional[Type[config.DBModel]]:
        """
        Get model by its name.

        :param model_name: Name of the model to find.
        :return: Model class if exists, None otherwise.
        """
        return next(
            (model for model in self.get_models() if model.__name__ == model_name), None
        )



from dataclasses import dataclass
from types import ModuleType
from typing import Set

from flask_sub_apps import config


@dataclass(frozen=True)
class SubApp:
    name: str
    module: ModuleType

    models: Set[config.DBModel]
    commands: Set[config.Command]
    routers: Set[config.Router]

    api_prefix: str = ""

    @property
    def module_name(self):
        return self.module.__name__

    def __str__(self):
        return self.module_name



import logging
from functools import reduce
from typing import Iterable, Any, List


def in_quotes(string: str) -> str:
    """Wrap string in double quotes."""
    return f'"{string}"'


def merge_lists(lists: Iterable[List[Any]]) -> List[Any]:
    """Merge an iterable of lists into a single list."""
    return reduce(lambda acc, arr: acc + arr, lists, [])


def get_logger() -> logging.Logger:
    """Get logger for FlaskSubApps."""
    return logging.getLogger("FlaskSubApps")



import importlib
import inspect
from functools import partial
from types import ModuleType
from typing import Union, Any, Callable, List

from flask_sub_apps import helpers, config, type_validation
from flask_sub_apps.models import SubApp

ValidateFunction = Callable[[Any], bool]


class ModuleLoader:
    """Loader of Flask sub-apps."""

    def __init__(
        self,
        is_valid_model: ValidateFunction = None,
        is_valid_command: ValidateFunction = None,
        is_valid_router: ValidateFunction = None,
    ):
        """
        Initialize loader with validators for models, commands and routers.
        Those functions will be used to filter out invalid items when loading modules.

        :param is_valid_model: function returning if obj is valid model class.
        :param is_valid_command: function returning if obj is valid command object.
        :param is_valid_router: function returning if obj is valid router object.
        """

        self.is_valid_model = is_valid_model or type_validation.is_valid_model
        self.is_valid_command = is_valid_command or type_validation.is_valid_command
        self.is_valid_router = is_valid_router or type_validation.is_valid_router

    def load_sub_app(self, module_name: str) -> SubApp:
        """
        Find and load a Flask sub-app by its module name.
        Method will try to load sub-app items from user defined keys in config.

        :param module_name: name of the module. e.g. 'backend.users'
        :return: initialized sub-app
        """
        module = self._import_module(module_name)
        if not module:
            raise ModuleNotFoundError(f"Module {module_name} not found")

        get_submodule_items = partial(self._get_submodule_items, module)
        models = set(
            filter(
                self.is_valid_model,
                get_submodule_items(submodule_name=config.MODELS_FILE),
            )
        )
        commands = set(
            filter(
                self.is_valid_command,
                get_submodule_items(submodule_name=config.COMMANDS_FILE),
            )
        )
        routers = set(
            filter(
                self.is_valid_router,
                get_submodule_items(submodule_name=config.ROUTERS_FILE),
            )
        )
        app_module = SubApp(
            name=self._get_short_module_name(module_name),
            module=module,
            models=models,
            commands=commands,
            routers=routers,
        )
        return app_module

    def _get_submodule_items(
        self,
        module: ModuleType,
        submodule_name: Union[str, List[str]],
    ) -> List[Any]:
        """
        Get all items from submodule of specified name.

        :param module: module to get submodule from.
        :param submodule_name: name of submodule to get items from. Can be a list of names.
        :return: list of items from submodule.
        """

        # If submodule_name is a list, then we need to iterate over submodules
        # and get all items from each of them and merge them into a single list.
        if isinstance(submodule_name, list):
            return helpers.merge_lists(
                self._get_submodule_items(module, name) for name in submodule_name
            )

        # Because file might not be imported we need to assemble
        # the name of the submodule and try to import it.
        submodule_name = f"{module.__name__}.{submodule_name}"
        submodule = self._import_module(submodule_name)
        if not submodule:
            return []

        return self._get_members_of_module(submodule)

    @staticmethod
    def _get_short_module_name(module: ModuleType) -> str:
        """
        Get short version of name of a module.
        It is used to generate name of the sub-app.

        :param module: module to get short name of.
        :return: short name of the module.
        """
        return module.split(".")[-1]

    @staticmethod
    def _import_module(module_name: str) -> ModuleType:
        """
        Import module by its name.

        :param module_name: name of the module.
        :return: imported module.
        """
        try:
            return importlib.import_module(module_name)
        except (ModuleNotFoundError, ImportError):
            return None

    @staticmethod
    def _get_members_of_module(module: ModuleType) -> List[Any]:
        """
        Get all members of a module.
        :param module: module to get members of.
        :return: list of members of the module.
        """
        return [self for name, self in inspect.getmembers(module)]



from typing import Iterable, Union, List

import flask
from flask import Flask

from flask_sub_apps import helpers, config
from flask_sub_apps.loader import ValidateFunction
from flask_sub_apps.models import SubApp
from flask_sub_apps.registry import AppRegistry

logger = helpers.get_logger()


class FlaskSubApps:
    def __init__(
        self,
        app: flask.Flask = None,
        installed_sub_apps: List[str] = None,
        models_file: Union[str, List[str]] = None,
        commands_file: Union[str, List[str]] = None,
        routers_file: Union[str, List[str]] = None,
        registry: AppRegistry = None,
        is_valid_model: ValidateFunction = None,
        is_valid_command: ValidateFunction = None,
        is_valid_router: ValidateFunction = None,
        model_class: type = None,
        router_class: type = None,
        command_class: type = None,
    ):
        """
        Initialize FlaskSubApps extension.

        :param app: Flask app to initialize FlaskSubApps with.
        :param installed_sub_apps: List of sub apps to install.
        :param models_file: Path to models file. List of paths is also accepted.
        :param commands_file: Path to commands module. List of paths is also accepted.
        :param routers_file: Path to routers module. List of paths is also accepted.
        :param registry: Instance of app registry to use.
        :param is_valid_model: Function to validate model.
        :param is_valid_command: Function to validate command.
        :param is_valid_router: Function to validate router.
        :param model_class: Model class to look for.
        :param router_class: Router class to look for.
        :param command_class: Command class to look for.
        """
        self._app = app
        self._registry: AppRegistry = registry or AppRegistry(
            is_valid_model=is_valid_model,
            is_valid_command=is_valid_command,
            is_valid_router=is_valid_router,
        )
        self._installed_sub_apps = installed_sub_apps or []
        self.init_config(
            models_file,
            commands_file,
            routers_file,
            model_class,
            router_class,
            command_class,
        )

        if app is not None:
            self.init_app(app)

    def init_app(self, app: Flask):
        """
        Initialize FlaskSubApps extension with Flask app,
        find and install all sub apps.

        :param app: Flask app to initialize FlaskSubApps with.
        """
        if self._app is app:
            logger.info("App is already initialized. Skipping.")
            return

        if self._app is not None:
            logger.warning(
                "FlaskSubApps is already initialized with another Flask app. "
                "Initializing with new Flask app."
            )
        if app is None:
            raise ValueError("Flask app is required to initialize FlaskSubApps.")

        self._app = app
        if not app.extensions:
            app.extensions = {}
        app.extensions[self.__class__.__name__] = self

        for module_name in self._installed_sub_apps:
            self._registry.register(module_name, app=self._app)

        logger.info(
            f"Registered [{', '.join(map(helpers.in_quotes, self._registry.get_sub_apps()))}] sub-apps."
        )

    @property
    def routers(self):
        """Get all registered routers."""
        return self._registry.get_routers()

    @property
    def commands(self):
        """Get all registered commands."""
        return self._registry.get_commands()

    @property
    def models(self):
        """Get all registered models."""
        return self._registry.get_models()

    def get_model(self, model_name: str):
        """
        Get model by its name.

        :param model_name: Name of the model to find.
        :return: Model class if exists, None otherwise.
        """
        return self._registry.get_model(model_name)

    def __iter__(self) -> Iterable[SubApp]:
        return iter(self._registry.get_sub_apps())

    @staticmethod
    def init_config(
        models_file: str,
        commands_file: str,
        routers_file: str,
        model_class: type,
        router_class: type,
        command_class: type,
    ):
        """
        Initialize config with user defined config files.
        """
        if models_file is not None:
            config.MODELS_FILE = models_file
        if commands_file is not None:
            config.COMMANDS_FILE = commands_file
        if routers_file is not None:
            config.ROUTERS_FILE = routers_file
        if model_class is not None:
            config.DBModel = model_class
        if router_class is not None:
            config.Router = router_class
        if command_class is not None:
            config.Command = command_class



# -*- coding: utf-8 -*-
from setuptools import setup, find_packages

try:
    long_description = open("README.rst").read()
except IOError:
    long_description = ""

setup(
    name="flask-sub-apps",
    version="0.1.3",
    description="Flask package for auto-loading sub-apps",
    license="MIT",
    author="Damian Ciftci",
    packages=find_packages(),
    install_requires=["flask", "flask-sqlalchemy", "click"],
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python",
        "Framework :: Flask",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)



from abc import ABC

from flask_sub_apps import config


def is_abstract_class(obj):
    """Check if it's abstract class by looking for ABC is in the object's __bases__ attribute."""
    try:
        return ABC in obj.__bases__
    except AttributeError:
        return False


def is_valid_model(obj):
    """Check if the object is a valid SQLAlchemy model class."""
    return isinstance(obj, config.DBModel)


def is_valid_command(obj):
    """
    Check if the object is a valid Flask command class.
    Click commands are also accepted.
    """
    return isinstance(obj, config.AppCommand)


def is_valid_router(obj):
    """Check if the object is a valid APIFlask blueprint class."""
    try:
        return isinstance(obj, config.Router)
    except TypeError:
        return False
